/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Mar 21, 2005
 *
 */
package com.sunlabs.archsim.gasp;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.hierarchy.Entity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author ib27688
 * 
 * A group of components that form a GasP branch.
 * Undefined is the function that selects what branch will be the
 * destionation of the input data.
 * The data copy function simply passes the data received on an 
 * input to the selected output branch. This function may need
 * to be overriden.
 *
 */
public class Branch extends Component {
	

	// default action name
	public static final String BACTION= "branchAction";

	
	/** the number of branches */
	protected int numBranches;

	/** the input for the data */
	protected InputTerminal in;
	/* acknowldegmnet that the input has been received */
	protected OutputTerminal ackIn;
		
	/** outputs for the data */
	protected OutputTerminal[] out;
	/** acknowldegment that the output has been received */
	protected InputTerminal[] ackOut;
	
	/** The branch action - to be initialized in a subclass
	 * because one needs to produce a non-abstract branch action
	 * with defined branch selection and possibly redefined data
	 * copy function.
	 */
	protected BranchAction branchAction;
	
	
	/** used in subclassess to process parameters */
	protected Branch(String name, Parameters params) {
		super(name, null);
	}

	
	public Branch(String name, int numOut, BranchAction ba) {
		super(name, null);
			
		build(numOut, ba);
		
	} // Branch

	
	/** default is two branches */
	protected Branch(String name, BranchAction ba) {
		this(name, 2, ba);
	} // Branch
	
	
	
	protected void build(int numOut, BranchAction ba) {
		// set the number of branches, must be at least 1
		numBranches= (numOut > 0)? numOut : 1;

		out= new OutputTerminal[numBranches];
		ackOut= new InputTerminal[numBranches];
		
		// make the terminals
		// input side first
		in= new InputTerminal(BranchAction.IN);
		addMember(in);
		
		ackIn= new OutputTerminal(BranchAction.ACKIN);
		addMember(ackIn);
		
		branchAction= ba;
		addMember(branchAction);
		
		// attach the input terminals
		branchAction.attachIn(in, ackIn);
		
		// now the output side
		for (int i= 0; i< numBranches; i++) {
			// make a pair of terminals
			OutputTerminal bout= new OutputTerminal(BranchAction.OUT 
										+ BranchAction.DELIMITER + i);
			addMember(bout);
			InputTerminal ackBout= new InputTerminal(BranchAction.ACKOUT 
										+ BranchAction.DELIMITER + i);
			addMember(ackBout);
			// add the new terminals to the vectors
			out[i] = bout;
			ackOut[i]= ackBout;
			
			// connect the terminals to the action
			branchAction.attachOut(out[i], ackOut[i], i);
		}		
	}
	
	/** set the same delay for all terminals */
	public void setDelayFromActionToState(Delay d) {
		in.setDelayFromActionToState(d);
		ackIn.setDelayFromActionToState(d);
		for (int i=0; i<numBranches; i++) {
			out[i].setDelayFromActionToState(d);
			ackOut[i].setDelayFromActionToState(d);
		}
	} // setDelayFromActionToState
	
	/** set the same delay for all terminals */
	public void setDelayFromStateToAction(Delay d) {
		in.setDelayFromStateToAction(d);
		ackIn.setDelayFromStateToAction(d);
		for (int i=0; i<numBranches; i++) {
			out[i].setDelayFromStateToAction(d);
			ackOut[i].setDelayFromStateToAction(d);
		}
	} // setDelayFromStateToAction
	
	/** accessor for the data input terminal */
	public InputTerminal getIn() {
		return in;
	} // getInput
	
	/** accessor for the ack-data input terminal */
	public OutputTerminal getAckIn() {
		return ackIn;
	} // getInput
	
	/** accessor for a data output terminal */
	public OutputTerminal getOut(int i) {
		if (i < 0 || i >= numBranches) return null;
		else return out[i];
	} // getInput
	
	/** accessor for an ack-data-out input terminal */
	public InputTerminal getAckOut(int i) {
		if (i < 0 || i >= numBranches) return null;
		else return ackOut[i];
	} // getInput
	
	public Action getAction() {
		return branchAction;
	} // getAction
	
	public int getNumBranches() {
		return numBranches;
	} // getNumBranches
	
	public boolean selfCheck() {
		boolean pass= super.selfCheck();
		pass= pass && (branchAction != null);
		if (!pass) {
			director.error("Branch " + getAlias() + ", id= " + getID()
					+ ", has null branchAction");
		}
		return pass;
	}
	
	public Terminal getTerminal(String name) {
		Terminal t=null;
		for (Entity e: members) {
			if (e instanceof Terminal) {
				Terminal tFound= (Terminal)e;
				if (tFound.getBaseName().equals(name)
						|| tFound.getName().equals(name)
						|| tFound.getBaseAlias().equals(name)
						|| tFound.getAlias().equals(name)) {
					t= tFound;
					break;
				}
			}
		}
		return t;
	}
	
} // Branch
